/*
 * Segment 6 Intelligente Bewertungsverfahren, SWE 6.1 Stauverlaufsanalyse
 * Copyright (C) 2007 BitCtrl Systems GmbH
 *
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 51
 * Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 *
 * Contact Information:
 * BitCtrl Systems GmbH
 * Weißenfelser Straße 67
 * 04229 Leipzig
 * Phone: +49 341-490670
 * mailto: info@bitctrl.de
 */

package de.bsvrz.ibv.anastau.bestimmungstau.stau;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.Comparator;
import java.util.List;
import java.util.Map.Entry;

import de.bsvrz.ibv.anastau.bestimmungstau.indikator.IFahrtRichtung;
import de.bsvrz.ibv.anastau.bestimmungstau.indikator.ISobIndikator;
import de.bsvrz.sys.funclib.debug.Debug;

/**
 * Verwaltung der Staus eines Straßenzuges, d.h. einer Menge von miteinander
 * verbundenen Stöerfallindikatoren.
 *
 * @author BitCtrl Systems GmbH, Peuker
 */
public class StauObjektVerwaltung {

	/**
	 * Comparator zum Ordnen von Stauobjekten. Stauobjekte werden nach ihrem
	 * Begin auf der Fahrtrichtung, auf der sie sich befinden geordnet.
	 */
	public static class StauMengenComparator implements Comparator<Stau> {

		/**
		 * Die Funktion ordnet Stauobjekte nach ihrem Beginn auf der Straße.
		 */
		@Override
		public int compare(final Stau o1, final Stau o2) {
			int result = 0;
			if (o1.getStrassenStart() < o2.getStrassenStart()) {
				result = -1;
			} else if (o1.getStrassenStart() > o2.getStrassenStart()) {
				result = 1;
			}
			return result;
		}
	}

	/** Logger für Debugaugaben. */
	private static final Debug LOGGER = Debug.getLogger();

	/** die Fahrtrichtung, für die die Staus verwaltet werden sollen. */
	private final IFahrtRichtung richtung;

	/**
	 * die Liste der Stauobjekte, die vom Datenverteiler entfernt werden können.
	 */
	private final Collection<IStau> entfernteStauobjekte = new ArrayList<>();

	/**
	 * die Menge der Staus, die innerhalb der Fahrtrichtung verwaltet werden.
	 */
	private final List<Stau> stauMenge = new ArrayList<>();

	/**
	 * Konstruktor.<br>
	 * Die Funktion erzeugt ein Verwaltungsobjekt für die angegebene
	 * Fahrtrichtung.
	 *
	 * @param fahrtRichtung
	 *            die Fahrtrichtung, deren Stauobjekte verwaltet werden sollen
	 */
	public StauObjektVerwaltung(final IFahrtRichtung fahrtRichtung) {
		richtung = fahrtRichtung;
	}

	/**
	 * ermittelt alle Staus, die innerhalb des angegebenen Bereichs innerhalb
	 * der Fahrtrichtung liegen.
	 *
	 * @param von
	 *            der Anfangswert des Bereiches, in dem Staus gesucht werden
	 * @param bis
	 *            der Endwert des Bereiches, in dem Staus gesucht werden
	 * @return die Liste der ermittelten Stauobjekte
	 */
	private Collection<Stau> ermittleStaus(final double von, final double bis) {
		final Collection<Stau> result = new ArrayList<>();
		for (final Stau stau : getStauMenge()) {
			final double stauStart = stau.getStrassenStart();
			final double stauEnde = stau.getStrassenEnde();
			if ((stauStart <= bis) && (stauEnde >= von)) {
				result.add(stau);
			}
		}
		return result;
	}

	/**
	 * ermittelt alle Staus, die gebildet werden, weil der übergebene Indikator
	 * beteiligt ist.
	 *
	 * @param ind
	 *            der Indikator
	 * @return die Liste der ermittelten Stauobjekte
	 */
	private Collection<Stau> ermittleStaus(final ISobIndikator ind) {
		final Collection<Stau> result = new ArrayList<>();
		if (ind != null) {
			for (final Stau stau : getStauMenge()) {
				if (stau.getIndikatoren().contains(ind)) {
					result.add(stau);
				}
			}
		}
		return result;
	}

	/**
	 * liefert eine sortierte Menge der Stauobjekte, der Fahrtrichtung die hier
	 * verwaltet wird.
	 *
	 * @return die sortierte Liste der Stauobjekte
	 */
	public List<Stau> getStauMenge() {
		Collections.sort(stauMenge, new StauMengenComparator());
		return stauMenge;
	}

	/**
	 * fügt der Stauobjektverwaltung einen neuen gestauten Bereich hinzu.<br>
	 * Der neue Bereich wird durch den übergebenen Störfallindikator
	 * beschrieben.<br>
	 * Die Funktion ermittelt alle Staus, die innerhalb des Bereiches des neu
	 * hinzugekommenen Störfallindikators liegen, wobei der Suchbereich um den
	 * parametrierten zulässigen Abstand zum zusammenführen von Stauobjekten
	 * erweitert wird. Alle gefundenen Staus werden zu einem Stauobjekt
	 * vereinigt. Der neue STauindikator wird für die Beschreibung des
	 * Staubereiches integriert.
	 *
	 * @param ind
	 *            der Indikator, der den neu hinzugekommenen gestauten Bereich
	 *            beschreibt.
	 */
	public final void neuerStauBereich(final ISobIndikator ind) {

		final double von = ind.getStauBeginn();
		final double bis = ind.getStauEnde();

		StauObjektVerwaltung.LOGGER.finer("Neuer gestauter Bereich von " + von + "m bis " + bis + " m");

		final Collection<Stau> stauListe = ermittleStaus(von, bis);
		Stau stau = null;
		if (stauListe.size() > 0) {
			for (final Stau oldStau : stauListe) {
				if (stau == null) {
					stau = oldStau;
				} else {
					stau.addIndikatoren(oldStau.getIndikatoren());
					entfernteStauobjekte.add(oldStau);
					stauMenge.remove(oldStau);
				}
			}
		} else {
			stau = new Stau(richtung);
			stauMenge.add(stau);
		}
		if (stau != null) {
			stau.addIndikator(ind);
		}
	}

	/**
	 * die Funktion fügt der Stauobjektverwaltung einen neuen ungestauten
	 * Bereich hinzu.<br>
	 * Der neue ungestaute Bereich wird durch den übergebenen Störfallindikator
	 * beschrieben.<br>
	 * Es werden alle Staus ermittelt, die innerhalb des betroffenen Bereiches
	 * liegen. Der neue ungestaute Bereich wird zur potentiellen Trennung des
	 * Staus an diesen übergeben.
	 *
	 * @param ind
	 *            der Störfallindikator, der den neuen ungestauten Bereich
	 *            repräsentiert.
	 * @param zeitpunkt
	 *            der Zeitpunkt, zu dem der Bereich betrachtet wird
	 */
	public final void neuerStaufreierBereich(final ISobIndikator ind, final long zeitpunkt) {

		boolean trennen = false;

		final double von = ind.getStauBeginn();
		final double bis = ind.getStauEnde();
		StauObjektVerwaltung.LOGGER.finer("Neuer ungestauter Bereich von " + von + "m bis " + bis + " m");

		final Collection<Stau> stauListe = ermittleStaus(ind);
		for (final Stau stau : stauListe) {
			Long zeit = null;
			switch (ind.getSituation()) {
			case UNDEFINIERT:
				zeit = stau.getTrenner().get(ind);
				if ((zeit != null) && (zeit < zeitpunkt)) {
					trennen = true;
				} else {
					stau.setTrenner(ind, zeitpunkt);
				}
				break;
			case VERKEHR:
				final float teilungsLaenge = ind.getStrasse().getParameter().getTeilungLaenge();
				final double luecke = stau.getStauLuecke(ind);
				if (luecke >= teilungsLaenge) {
					trennen = true;
				} else {
					final long verzoegerung = ind.getStrasse().getParameter().getTeilungsVerzoegerung();
					zeit = stau.getTrenner().get(ind);
					if ((zeit != null) && ((zeitpunkt - zeit) > verzoegerung)) {
						trennen = true;
					} else {
						stau.setTrenner(ind, zeitpunkt);
					}
				}
				break;
			case STAU:
				stau.entferneTrenner(ind);
				break;
			default:
				break;
			}

			if (trennen) {
				trenneStau(stau, ind);
			}
		}
	}

	/**
	 * berechnet die aktuelle Prognoseverlängerung für alle verwalteten
	 * Stauobjekte.
	 */
	public void prognoseVerlaengerungBerechnen() {
		for (final Stau stau : stauMenge) {
			stau.aktualisierePrognoseVerlaengerung();
		}
	}

	/**
	 * prüft die aktuelle Prognoseverlängerung und verkürzt diese unter
	 * bestimmtem Bedingungen.
	 */
	public void prognoseVerlaengerungPruefen() {
		for (final Stau stau : stauMenge) {
			final double start = stau.getStrassenStart();
			final double ende = stau.getStrassenEnde();
			final double maxVerl = richtung.getMaxPrognoseLaenge(start, ende);
			stau.setPrognoseVerlaengerung(Math.min(stau.getPrognoseVerlaengerung(), maxVerl));
		}
	}

	/**
	 * Die Funktion entfernt zuerset alle zum Löschen vorgemerkten Stauobjekte
	 * und sichert danach alle noch bestehenden, dabei werden je nach Bedarf
	 * neue Objekte angelegt oder die bestehenden mit altuellen Daten versehen.
	 *
	 * @param zeitpunkt
	 *            der Zeitpunkt für den die Sicherung der Stauobjekte erfolgt
	 */
	public final void sichern(final long zeitpunkt) {
		for (final IStau stau : entfernteStauobjekte) {
			stau.entfernen();
			stauMenge.remove(stau);
		}
		entfernteStauobjekte.clear();
		for (final Stau stau : getStauMenge()) {
			if (stau.getIndikatoren().size() <= 0) {
				entfernteStauobjekte.add(stau);
			} else {
				stau.sichern(zeitpunkt);
			}
		}
	}

	/**
	 * trennt ein Stauobjekt an der Stelle, an der sich der übergebene Indikator
	 * befindet. Liegt der Indikator innerhalb des Staus, entstehen zwei neue
	 * Stauohbjekte, liegt der Indikator am Ende oder am Anfang des Stauobjekts,
	 * wird dieses lediglich entsprechend verkürzt.
	 *
	 * @param stau
	 *            der Stau, der getrennt werden soll
	 * @param indikator
	 *            der Indikator, an dem die Trennung erfolgen soll
	 * @return <code>true</code>, wenn neue Stauobjekte entstanden sind
	 */
	private boolean trenneStau(final Stau stau, final ISobIndikator indikator) {
		boolean result = false;
		final double von = indikator.getStauBeginn();
		final Stau neuerStau = new Stau(richtung);
		for (final ISobIndikator stauIndikator : stau.getIndikatoren()) {
			if (stauIndikator.equals(indikator)) {
				stau.entferneIndikator(stauIndikator);
				break;
			}
			neuerStau.addIndikator(stauIndikator);
			stau.entferneIndikator(stauIndikator);
			neuerStau.setPrognoseVerlaengerung(von - neuerStau.getStrassenEnde());
		}
		if (neuerStau.getIndikatoren().size() > 0) {
			stauMenge.add(neuerStau);
			result = true;
		}
		if (stau.getIndikatoren().size() <= 0) {
			stauMenge.remove(stau);
			entfernteStauobjekte.add(stau);
		}
		return result;
	}

	/**
	 * alle für eine Trennung vorgemerkten Stauobjekte werden getrennt, wenn die
	 * parametrierte Timeout-Zeit abgelaufen ist.
	 *
	 * @param zeitpunkt
	 *            der Zeitpunkt für den die verzögerte Trennung erfolgen soll
	 */
	public final void verzoegerteTrennung(final long zeitpunkt) {
		final long verzoegerung = richtung.getStrasse().getParameter().getTeilungsVerzoegerung();

		List<Stau> liste = new ArrayList<>(getStauMenge());
		while (liste.size() > 0) {
			final Stau stau = liste.get(0);
			liste.remove(stau);
			for (final Entry<ISobIndikator, Long> eintrag : stau.getTrenner().entrySet()) {
				boolean trennen = false;
				switch (eintrag.getKey().getSituation()) {
				case UNDEFINIERT:
					if (zeitpunkt > eintrag.getValue()) {
						trennen = true;
					}
					break;
				case VERKEHR:
					if ((zeitpunkt - eintrag.getValue()) > verzoegerung) {
						trennen = true;
					}
					break;
				case STAU:
					stau.entferneTrenner(eintrag.getKey());
					break;
				default:
					break;
				}
				if (trennen) {
					if (trenneStau(stau, eintrag.getKey())) {
						liste = new ArrayList<>(getStauMenge());
					}
				}
			}
		}
	}

	/**
	 * fasst Stauobjekte zusammen.
	 */
	public void zusammenfassen() {
		final Collection<Stau> ergebnisListe = new ArrayList<>();
		Stau aktuellerStau = null;
		final double abstand = richtung.getStrasse().getParameter().getMaxSiAbstand();

		for (final Stau stau : getStauMenge()) {
			if (aktuellerStau == null) {
				aktuellerStau = stau;
				ergebnisListe.add(aktuellerStau);
			} else {
				final double ende = aktuellerStau.getStrassenEnde();
				final double start = stau.getStrassenStart();
				if ((start - ende) < abstand) {
					aktuellerStau.addIndikatoren(stau.getIndikatoren());
					aktuellerStau.setPrognose(stau.getPrognose());
					aktuellerStau.setPrognoseVerlaengerung(stau.getPrognoseVerlaengerung());
					entfernteStauobjekte.add(stau);
				} else {
					aktuellerStau = stau;
					ergebnisListe.add(aktuellerStau);
				}
			}
		}

		if (ergebnisListe.size() != stauMenge.size()) {
			stauMenge.clear();
			stauMenge.addAll(ergebnisListe);
		}

		StauObjektVerwaltung.LOGGER.finer("Stauliste zusammengefasst: " + ergebnisListe.size() + "Stau(s)");
	}
}
