/*
 * Segment 6 Intelligente Bewertungsverfahren, SWE 6.1 Stauverlaufsanalyse
 * Copyright (C) 2007 BitCtrl Systems GmbH
 *
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 51
 * Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 *
 * Contact Information:
 * BitCtrl Systems GmbH
 * Weißenfelser Straße 67
 * 04229 Leipzig
 * Phone: +49 341-490670
 * mailto: info@bitctrl.de
 */

package de.bsvrz.ibv.anastau.progstau;

import java.text.DateFormat;
import java.util.Date;

import de.bsvrz.ibv.anastau.anastau.StauverlaufsAnalyse;
import de.bsvrz.ibv.anastau.anastau.SvaMessage;
import de.bsvrz.ibv.anastau.bestimmungstau.indikator.SobMessQuerschnitt;
import de.bsvrz.ibv.anastau.bestimmungstau.indikator.SobStrassenKnoten;
import de.bsvrz.ibv.anastau.bestimmungstau.indikator.SobStrassenTeilSegment;
import de.bsvrz.ibv.anastau.bestimmungstau.stau.IStau;
import de.bsvrz.ibv.anastau.progstau.parameter.ISvpParameterObserver;
import de.bsvrz.sys.funclib.bitctrl.daf.BetriebsmeldungDaten;
import de.bsvrz.sys.funclib.bitctrl.daf.LogTools;
import de.bsvrz.sys.funclib.bitctrl.modell.verkehr.objekte.MessQuerschnittAllgemein;
import de.bsvrz.sys.funclib.bitctrl.modell.verkehr.objekte.StoerfallIndikator;
import de.bsvrz.sys.funclib.bitctrl.modell.verkehr.objekte.StrassenTeilSegment;
import de.bsvrz.sys.funclib.bitctrl.modell.verkehr.parameter.PdFundamentalDiagramm;
import de.bsvrz.sys.funclib.bitctrl.util.UndefiniertException;
import de.bsvrz.sys.funclib.debug.Debug;

/**
 * Implemtierung einer Klasse zur Bereitstellung der Engpasskapazität eines
 * Stauobjekts.
 *
 * @author BitCtrl Systems GmbH, Peuker
 */
public class SvpKapazitaet implements ISvpKapazitaet {

	/**
	 * Logger für Debugausgaben.
	 */
	private static final Debug LOGGER = Debug.getLogger();

	/** der für die Kapazität ermittelte Wert. */
	private Number wert;

	/** der für die Bestimmung der Kapazität verwendete Indikator. */
	private StoerfallIndikator usedIndikator;

	/**
	 * erzeugt eine Instanz zur Ermittlung der Engpasskapazit&auml;t für einen
	 * Stau.<br>
	 *
	 * @param stau
	 *            der Stau, für den die Engpasskapazit&auml;t ermittelt werden
	 *            soll
	 * @param parameter
	 *            die Parameter der Stauverlaufsprognose
	 */
	public SvpKapazitaet(final IStau stau, final ISvpParameterObserver parameter) {
		final SobMessQuerschnitt mqVorDemStau = stau.getMqDanach();
		MessQuerschnittAllgemein stauMq = null;

		if (mqVorDemStau != null) {
			stauMq = mqVorDemStau.getIndikator();
			usedIndikator = stauMq;
			wert = StauverlaufsPrognose.getMqQbWert(stauMq);
		}

		if (StauverlaufsAnalyse.getOptionen().benutzeAnschluesseNachDemStau()) {
			// Zu- und Abfahrten einkalkulieren
			if ((wert != null) && (usedIndikator != null)) {
				for (final SobStrassenKnoten knoten : stau.ermittleAsAmAnfang(stauMq)) {
					for (final MessQuerschnittAllgemein mq : knoten.getAbfahrtMq()) {
						final Number mqWert = StauverlaufsPrognose.getMqQbWert(mq);
						if (mqWert != null) {
							wert = wert.doubleValue() + mqWert.doubleValue();
						}
					}
					for (final MessQuerschnittAllgemein mq : knoten.getZufahrtMq()) {
						final Number mqWert = StauverlaufsPrognose.getMqQbWert(mq);
						if (mqWert != null) {
							wert = wert.doubleValue() - mqWert.doubleValue();
						}
					}
				}
			}
		}

		// Kapazität aus den Q0-Werten des Fundamentaldiagramms ermitteln
		if (wert == null) {
			final SobStrassenTeilSegment stsVorDemStau = stau.getStsVorDemStau();
			final SobStrassenTeilSegment stsAmEnde = stau.getLetztesStsImStau();
			if (stsVorDemStau != null) {
				wert = getQ0Wert(stsVorDemStau.getIndikator());
				usedIndikator = stsVorDemStau.getIndikator();
			}
			if (stsAmEnde != null) {
				final Number vergleichsWert = getQ0Wert(stsAmEnde.getIndikator());
				if ((wert == null)
						|| ((vergleichsWert != null) && (vergleichsWert.doubleValue() < wert.doubleValue()))) {
					wert = vergleichsWert;
					usedIndikator = stsAmEnde.getIndikator();
				}
			}

			if (wert != null) {
				wert = (wert.doubleValue() * parameter.getFaktorQ0()) / 100.0;
			}
		}
	}

	@Override
	public final double getKapazitaet(final long zeitPunkt) throws UndefiniertException {
		if (wert == null) {
			throw new UndefiniertException("Es existiert kein Kapazitätswert für den Zeitpunkt: "
					+ DateFormat.getDateTimeInstance().format(new Date(zeitPunkt)));
		}
		SvpKapazitaet.LOGGER.finer("Kapazität für " + usedIndikator + " um "
				+ DateFormat.getDateTimeInstance().format(new Date(zeitPunkt)) + " = " + wert.doubleValue());
		return wert.doubleValue();
	}

	/**
	 * ermittelt den Q0-Wert des Fundamentaldiagramms des &uuml;bergebenen
	 * Straßenteilsegments.
	 *
	 * @param sts
	 *            das Straßenteilsegment
	 * @return der ermittelte Wert oder <code>null</code>, wenn keiner ermittelt
	 *         werden konnte.
	 */
	private Number getQ0Wert(final StrassenTeilSegment sts) {

		Number result = null;
		final PdFundamentalDiagramm.Daten daten = sts.getParameterDatensatz(PdFundamentalDiagramm.class).getDatum();
		if (daten != null) {
			result = daten.getQ0();
		}

		if (result == null) {
			LogTools.log(SvpKapazitaet.LOGGER, new BetriebsmeldungDaten(sts.getSystemObject()),
					SvaMessage.WarningKeinQ0Wert, sts);
		}

		return result;
	}
}
